﻿function Complex(re, im) {
	if ( !(this instanceof Complex) ) //allow calling 'Complex()' without 'new'
		return new Complex(re, im);
	
	if (typeof re === 'object')
		re = re.valueOf();
		
	if (!Number.isFinite(re))
		throw "Complex: the real part must be a finite number";
	
	this._re = re;
	
	if (typeof im === 'undefined') {
		this._im = 0;
		return this;
	}
	
	if (typeof im === 'object')
		im = im.valueOf();
	
	if (!Number.isFinite(im))
		throw "Complex: the imaginary part must be a finite number";
	
	this._im = im;
	return this;
};

Complex._create = function(re, im) {
	var c = Object.create(Complex.prototype);
	c._re = re;
	c._im = im;
	return c;
}

Complex.polar = function(r, θ) {
	if (typeof r === 'object')
		r = r.valueOf();
	
	if (typeof θ === 'object')
		θ = θ.valueOf();
	
	if (!Number.isFinite(r) || !Number.isFinite(θ))
		throw "Complex.polar: invalid arguments (finite numbers expected)";

	return Complex._create(r * Math.cos(θ), r * Math.sin(θ));
};

Complex.prototype.toStr = function(fn) {
	var re = this._re;
	var im = this._im;
	
	if (im === 0)
		return fn(re);
	
	if (re === 0)
		return fn(im) + 'i';
	
	return fn(re) + (im >= 0 ? '+' : '') + fn(im) + 'i';
};

Complex.prototype.toString = function() {
	return this.prec();
};

Complex.prototype.real = function() {
	return this._re;
};

Complex.prototype.imag = function() {
	return this._im;
};

Complex.prototype.add = function(x) {
	if (Number.isNumber(x))
		return Complex._create(this._re + x, this._im);
	
	if (x instanceof Complex)
		return Complex._create(this._re + x._re, this._im + x._im);
		
	throw "Complex.add: invalid type of argument (number or complex number expected)";
};

Complex.prototype.sub = function(x) {
	if (Number.isNumber(x))
		return Complex._create(this._re - x, this._im);
	
	if (x instanceof Complex)
		return Complex._create(this._re - x._re, this._im - x._im);
		
	throw "Complex.sub: invalid type of argument (number or complex number expected)";
};

Complex.prototype.mul = function(x) {
	if (Number.isNumber(x))
		return Complex._create(this._re * x, this._im * x);
	
	if (x instanceof Complex)
		return Complex._create(this._re * x._re - this._im * x._im, this._re * x._im + this._im * x._re);
		
	throw "Complex.mul: invalid type of argument (number or complex number expected)";
};

Complex.prototype.div = function(x) {
	if (Number.isNumber(x))
		return Complex._create(this._re / x, this._im / x);
	
	if (x instanceof Complex) {
		var m = x._re * x._re + x._im * x._im;
		var re = (this._re * x._re + this._im * x._im) / m;
		var im = (this._im * x._re - this._re * x._im) / m;
		
		return Complex._create(re, im);
	}
		
	throw "Complex.div: invalid type of argument (number or complex number expected)";
};

Complex.prototype.mulIm = function() {
	return Complex._create(-this._im, this._re);
};

Complex.prototype.divIm = function() {
	return Complex._create(this._im, -this._re);
};

Complex.prototype.abs = function() {
	return Math.sqrt( this._re * this._re + this._im * this._im );
};

Complex.prototype.arg = function() {
	return Math.atan2(this._im, this._re);
};

Complex.prototype.cc = function() {
	return Complex._create(this._re, -this._im); //complex conj.
};

Complex.prototype.exp = function() {
	var a = Math.exp(this._re);
	return new Complex( a * Math.cos(this._im), a * Math.sin(this._im) );
};

Complex.prototype.log = function() {
	return new Complex( Math.log(this.abs()), this.arg() ); //principal value
};

Complex.prototype.pow = function(x) {
	return this.log().mul(x).exp();
};

Complex.prototype.sq = function() {
	return Complex._create(this._re * this._re - this._im * this._im, 2 * this._re * this._im);
};

Complex.prototype.sqrt = function() {
	return this.pow(0.5);
};

Complex.prototype.neg = function() {
	return Complex._create(-this._re, -this._im);
};

Complex.prototype.sin = function() {
	var a = this.mulIm().exp();
	var b = this.divIm().exp();
  	return b.sub(a).mul(Complex(0, 0.5));
};

Complex.prototype.cos = function() {
	var a = this.mulIm().exp();
	var b = this.divIm().exp();
  	return a.add(b).mul(0.5);
};

Complex.prototype.tan = function() {
	var a = this.mulIm().exp();
	var b = this.divIm().exp();
  	return b.sub(a).mul(Complex(0, 1)).div(a.add(b));
};

Complex.prototype.sinh = function() {
	return this.exp().sub(this.neg().exp()).mul(0.5);
};

Complex.prototype.cosh = function() {
	return this.exp().add(this.neg().exp()).mul(0.5);
};

Complex.prototype.tanh = function() {
	var a = this.exp();
	var b = this.neg().exp();
	return a.sub(b).div(a.add(b));
};

Complex.prototype.asinh = function() {
	return this.sq().add(1).sqrt().add(this).log();
};

Complex.prototype.acosh = function() {
	return this.add(1).sqrt().mul(this.sub(1).sqrt()).add(this).log();
};

Complex.prototype.atanh = function() {
	return this.add(1).log().sub(this.neg().add(1).log()).mul(0.5);
};

Complex.prototype.atan = function() {
	return this.mulIm().atanh().divIm();
};

Complex.prototype.asin = function() {
	return this.mulIm().asinh().divIm();
};

Complex.prototype.equals = function(x) {
	if (Number.isNumber(x))
		return this._re === x && this._im === 0;
	
	if (x instanceof Complex)
		return this._re === x._re && this._im === x._im;
		
	throw "Complex.equal: undefined operation";
};








