﻿function Matrix() {
	if ( !(this instanceof Matrix) ) { //allow calling 'Matrix()' without 'new'
		var obj = Object.create(Matrix.prototype);
		Matrix.apply(obj, arguments);
		return obj;
	}

	if (arguments.length === 1 && Matrix.isValidArray(arguments[0])) {
		this._data = arguments[0];
		return this;
	}
	
	if (arguments.length === 3 && Number.isNumber(arguments[0]) && Number.isNumber(arguments[1]) && Function.isFunction(arguments[2])) {
		var rows = Math.floor(arguments[0]);
		var cols = Math.floor(arguments[1]);
		if (rows < 1 || cols < 1)
			throw "Matrix: invalid number of rows or columns";
		
		var f = arguments[2];
		this._data = [];
		for (var r = 0; r < rows; r++) {
			var row = [];
			for (var c = 0; c < cols; c++)
				row.push( f(r,c) );
			this._data.push(row);
		}
		
		return this;
	}
	
	throw "Matrix: invalid combination of arguments";
}

Matrix._create = function(data) {
	var m = Object.create(Matrix.prototype);
	m._data = data;
	return m;
}

Matrix.prototype.array = function() {
	return this._data.clone();
}

Matrix.prototype.rows = function() {
	return this._data.length;
}

Matrix.prototype.cols = function() {
	return this._data[0].length;
}

/* test for an array that would be suitable for Matrix */
Matrix.isValidArray = function(A) {
	if (!Array.isArray(A) || A.length === 0 || !Array.isArray(A[0]))
		return false;
	
	var len = A[0].length;
	return A.every(function(row){
		return Array.isArray(row) && row.length === len && row.every(Number.isFinite);
	});
}

Matrix.prototype.equals = function(x) {
	return (x instanceof Matrix) && x._data.equals(this._data);
}

Matrix.Unit = function(dim) {
	if (typeof dim === 'object')
		dim = dim.valueOf();
	
	if (!Number.isFinite(dim) || dim < 1 || Math.floor(dim) !== dim)
		throw "Matrix.Unit: invalid dimension (must be integer >= 1)";
	
	var data = [];
	for (var i = 0; i < dim; i++) {
		var row = Array.const(dim, 0);
		row[i] = 1;
		data.push(row);
	}
	
	return Matrix._create(data);
}

Matrix.Vander = function(x, n) {
	if (x instanceof Vector)
		x = x._data;
	else if (!Vector.isValidArray(x))
		throw "Matrix.Vander: invalid array";

	var m = x.length;
	var data = [];
	for (var i = 0; i < m; i++) {
		var row = [];
		for (var j = 0; j < n; j++)
			row.push(Math.pow(x[i], j));
		data.push(row);
	}
	
	return Matrix._create(data);
}

Matrix._maxRowCount = 16;
Matrix._maxColCount = 8;

Matrix.prototype.toStr = function(fn) {
	var data = this._data;
	var m = data.length;
	var n = data[0].length;
	if (m > Matrix._maxRowCount)
		data = [ data[0], data[1], data[m-2], data[m-1] ];
	
	var f;
	if (n <= Matrix._maxColCount)
		f = function(row) { return row.map(fn); };
	else
		f = function(row) { return [fn(row[0]), fn(row[1]), '...', fn(row[n-2]), fn(row[n-1])]; };
	
	var res = data.map(f);
	var rows = res.length;
	var cols = res[0].length;
	for (var c = 0; c < cols; c++) {
		var w = 0;
		for (var r = 0; r < rows; r++)
			if (res[r][c].length > w)
				w = res[r][c].length;
		for (var r = 0; r < rows; r++)
			res[r][c] = res[r][c].pad(w);
	}
	
	res = res.map(function(row) {
		return "[" + row.join(", ") + "]";
	});
	
	if (m <= Matrix._maxRowCount)
		return "[" + res.join("\n ") + "]";
	
	return "[" + res.slice(0,2).join("\n ") + "\n ...\n " + res.slice(2).join("\n ") + "]";
}

Matrix.prototype.toString = function() {
	return this.fixed(3);
}

Matrix.prototype.slice = function(ri, rj, ci, cj) {
	var data = this._data.slice(ri, rj);
	if (data.length === 0)
		throw "Matrix.slice: invalid range of row indices";
	for (var k = 0; k < data.length; k++) {
		data[k] = data[k].slice(ci, cj);
		if (data[k].length === 0)
			throw "Matrix.slice: invalid range of column indices";
	}
	
	return Matrix._create(data);
}

Matrix.prototype.rows = function(ri, rj) {
	return this.slice(ri, rj);
}

Matrix.prototype.cols = function(ci, cj) {
	return this.slice(undefined, undefined, ci, cj);
}

Matrix.prototype.add = function(x) {
	var a = this._data;
	
	if (x instanceof Matrix) {
		x = x._data;
		if (a.length !== x.length || a[0].length !== x[0].length)
			throw "Matrix.add: size mismatch";
			
		return Matrix._create( a.add(x) );
	}
	
	if (Number.isNumber(x))
		return Matrix._create( a.add(x) );
	
	throw "Matrix.add: invalid type of argument (matrix or number expected)";
}

Matrix.prototype.sub = function(x) {
	var a = this._data;
	
	if (x instanceof Matrix) {
		x = x._data;
		if (a.length !== x.length || a[0].length !== x[0].length)
			throw "Matrix.sub: size mismatch";
			
		return Matrix._create( a.sub(x) );
	}
	
	if (Number.isNumber(x))
		return Matrix._create( a.sub(x) );
	
	throw "Matrix.sub: invalid type of argument (matrix or number expected)";
}

Matrix.prototype._mulArray = function(x) {
	var a = this._data;
	var n = a[0].length;
	if (n !== x.length)
		throw "Matrix.mul: size mismatch";
	
	/* size of the product matrix */
	var rows = a.length;
	var cols = x[0].length;
		
	var p = []; //product array
	for (var r = 0; r < rows; r++) {
		var row = [];
		for (var c = 0; c < cols; c++) {
			var val = 0;
			for (var i = 0; i < n; i++)
				val += a[r][i] * x[i][c];
			row.push(val);
		}
		p.push(row);
	}
	
	return Matrix._create(p);
}

Matrix.prototype.mul = function(x) {
	if (x instanceof Matrix)
		return this._mulArray(x._data);
	
	if (x instanceof Vector) {
		x = x._data;
		var a = this._data;
		var cols = a[0].length;
		var rows = a.length;
		if (cols !== x.length)
			throw "Matrix.mul: size mismatch";
			
		var data = [];
		for (var r = 0; r < rows; r++) {
			var val = 0;
			for (var i = 0; i < cols; i++)
				val += a[r][i] * x[i];
			data.push(val);
		}
		
		return Vector._create(data);
	}
	
	if (Number.isNumber(x))
		return Matrix._create( this._data.mul(x) );
	
	if (Matrix.isValidArray(x))
		return this._mulArray(x);
	
	throw "Matrix.mul: invalid type of argument (matrix, vector or number expected)";
}

Matrix.prototype.div = function(x) {
	if (Number.isNumber(x))
		return Matrix._create( this._data.div(x) );
	
	throw "Matrix.div: invalid type of argument (number expected)";
}

Matrix.prototype.trans = function() {
	var a = this._data;
	var rows = a[0].length;
	var cols = a.length;
	var data = [];
	for (var r = 0; r < rows; r++) {
		var row = [];
		for (var c = 0; c < cols; c++)
			row.push(a[c][r]);
		data.push(row);
	}
	
	return Matrix._create(data);
}

Matrix.prototype.addrows = function(x) {
	if (x instanceof Matrix)
		x = x._data;
	else if (!Matrix.isValidArray(x))
		throw "Matrix.addrows: undefined operation";
	
	var a = this._data;
	if (a[0].length !== x[0].length)
		throw "Matrix.addrows: size mismatch";

	a.extend(x);
	return this;
}

Matrix.prototype.addcols = function(x) {
	if (x instanceof Matrix)
		x = x._data;
	else if (!Matrix.isValidArray(x))
		throw "Matrix.addcols: undefined operation";
	
	var a = this._data;
	if (a.length !== x.length)
		throw "Matrix.addcols: size mismatch";

	for (var i = 0; i < a.length; i++)
		a[i].extend(x[i]);

	return this;
}

Matrix.prototype.isSquare = function() {
	var a = this._data;
	return a.length === a[0].length;
}

Matrix.prototype.clone = function() {
	return Matrix._create(this._data.clone());
}

Matrix.prototype.det = function() {
	var A = this._data.clone();
	var n = A.length;
	var D = 1;

	for (var i = 0; i < n; i++) { 
		var m = abs(A[i][i]); //find row with maximum element
		var r = i;
		for (var k = i+1; k < n; k++) { 
			if (Math.abs(A[k][i]) > m) {
				m = abs(A[k][i]);
				r = k;
			}
		}
		
		if (m === 0)
			return 0;
		
		if (r !== i) {
			var tmp = A[r]; //swap rows
			A[r] = A[i];
			A[i] = tmp;
			D *= -1;
		}
		
		D *= A[i][i];
		for (var k = i+1; k < n; k++) { 
			var c = A[k][i] / A[i][i];
			for (var j = i + 1; j < n; j++)
				A[k][j] = ( i === j ? 0 : A[k][j] - c * A[i][j] );
		}
	}

	return D;
}

Matrix.prototype.inv = function() {
	if (!this.isSquare())
		throw "Matrix.inv: matrix must be square";
		
	var n = this._data.length;
	var M = this.clone().addcols( Matrix.Unit(n) );
	var A = M._data;

	for (var i = 0; i < n; i++) { 
		var m = abs(A[i][i]); //find row with maximum element
		var r = i;
		for (var k = i+1; k < n; k++) { 
			if (Math.abs(A[k][i]) > m) {
				m = abs(A[k][i]);
				r = k;
			}
		}
		
		if (m === 0)
			throw "Matrix.inv: no solution found (singular matrix)";
		
		var tmp = A[r]; //swap rows
		A[r] = A[i];
		A[i] = tmp;
		
		for (var k = i+1; k < n; k++) { 
			var c = A[k][i] / A[i][i];
			for (var j = i; j < 2*n; j++)
				A[k][j] = ( i === j ? 0 : A[k][j] - c * A[i][j] );
		}
	}
	
	for (var i = n-1; i >= 0; i--) {
		var c = 1 / A[i][i];
		A[i][i] = 1;
		for (var j = i + 1; j < 2*n; j++)
			A[i][j] *= c;
	}
	
	for (var i = n-1; i > 0; i--)
		for (var k = i-1; k >= 0; k--) {
			var c = A[k][i];
			//A[k][i] = 0;
			for (var j = n; j < 2*n; j++) //for (var j = i+1; j < 2*n; j++)
				A[k][j] -= c * A[i][j];
		}
	
	return M.cols(n, 2*n);
}

Matrix.lsolve = function(A,b) {
	if (A instanceof Matrix)
		A = A._data;
	else if (!Matrix.isValidArray(A))
		throw "Matrix.lsolve: array A is invalid";
	
	if (A.length !== A[0].length)
		throw "Matrix.lsolve: matrix A must be square";
		
	if (b instanceof Vector)
		b = b._data;
	else if (!Vector.isValidArray(b))
		throw "Matrix.lsolve: array b is invalid";
	
	if (A.length !== b.length)
		throw "Matrix.lsolve: size mismatch";
	
	A = A.clone();
	for (var i = 0; i < A.length; i++) 
        A[i].push( b[i] );
	
	var n = A.length;

	for (var i = 0; i < n; i++) { 
		var m = abs(A[i][i]); //find row with maximum element
		var r = i;
		for (var k = i+1; k < n; k++) { 
			if (Math.abs(A[k][i]) > m) {
				m = abs(A[k][i]);
				r = k;
			}
		}
		
		if (m === 0)
			throw "Matrix.lsolve: no solution found (singular matrix)";
		
		if (r !== i) {
			var tmp = A[r]; //swap rows
			A[r] = A[i];
			A[i] = tmp;
		}
		
		for (var k = i+1; k < n; k++) { 
			var c = A[k][i] / A[i][i];
			for (var j = i; j < n + 1; j++)
				A[k][j] = ( i === j ? 0 : A[k][j] - c * A[i][j] );
		}
	}

	var x = new Array(n);
	for (var i = n-1; i >= 0; i--) { 
		x[i] = A[i][n] / A[i][i];
		for (var k = i-1; k >= 0; k--)
			A[k][n] -= A[k][i] * x[i];
	}
	
	return x;
}

function Vector() {
	if ( !(this instanceof Vector) ) { //allow calling 'Vector()' without 'new'
		var obj = Object.create(Vector.prototype);
		Vector.apply(obj, arguments);
		return obj;
	}

	if (arguments.length === 0)
		return this;

	if (arguments.length === 1 && Array.isArray(arguments[0])) {
		if (!Vector.isValidArray(arguments[0]))
			throw "Vector: invalid array";
		
		this._data = arguments[0];
		return this;
	}
	
	if (arguments.length === 2 && Number.isNumber(arguments[0]) && Function.isFunction(arguments[1])) {
		var len = Math.floor(arguments[0]);
		if (len < 1)
			throw "Vector: invalid number of elements";
		
		var f = arguments[1];
		this._data = [];
		for (var i = 0; i < len; i++)
			this._data.push( f(i) );
		
		return this;
	}
	
	throw "Vector: invalid combination of arguments";
}

Vector._create = function(data) {
	var v = Object.create(Vector.prototype);
	v._data = data;
	return v;
}

Vector.prototype.toStr = function(fn) {
	var data = this._data;
	var n = data.length;
	if (n <= Matrix._maxColCount)
		return "[" + data.map(fn).join(", ") + "]";
	
	return "[" + fn(data[0]) + ", " + fn(data[1]) + ", ..., " + fn(data[n-2]) + ", " + fn(data[n-1]) + "]";
}

Vector.prototype.toString = function() {
	return this.fixed(3);
}

/* test for an array that would be suitable for Vector */
Vector.isValidArray = function(A) {
	return Array.isArray(A) && A.length !== 0 && A.every(Number.isFinite);
}

Vector.prototype.dot = function(x) {
	if (x instanceof Vector)
		x = x._data;
	else if (!Vector.isValidArray(x))
		throw "Vector.dot: invalid type of argument (vector expected)";

	var a = this._data;
	if (a.length !== x.length)
		throw "Vector.dot: expected arrays of equal length";

	var res = 0;
	for (var i = 0; i < a.length; i++)
		res += a[i] * x[i];
	
	return res;
}

Vector.prototype.add = function(x) {
	var a = this._data;
	
	if (x instanceof Vector) {
		x = x._data;
		if (a.length !== x.length)
			throw "Vector.add: size mismatch";
			
		return Vector._create( a.add(x) );
	}
	
	if (Number.isNumber(x))
		return Vector._create( a.add(x) );
	
	if (Vector.isValidArray(x)) {
		if (a.length !== x.length)
			throw "Vector.add: size mismatch";
			
		return Vector._create( a.add(x) );
	}
	
	throw "Vector.add: invalid type of argument (vector or number expected)";
}

Vector.prototype.sub = function(x) {
	var a = this._data;
	
	if (x instanceof Vector) {
		x = x._data;
		if (a.length !== x.length)
			throw "Vector.sub: size mismatch";
			
		return Vector._create( a.sub(x) );
	}
	
	if (Number.isNumber(x))
		return Vector._create( a.sub(x) );
	
	if (Vector.isValidArray(x)) {
		if (a.length !== x.length)
			throw "Vector.add: size mismatch";
			
		return Vector._create( a.sub(x) );
	}
	
	throw "Vector.sub: invalid type of argument (vector or number expected)";
}

Vector.prototype.mul = function(x) {
	if (Number.isNumber(x))
		return Vector._create( this._data.mul(x) );
		
	throw "Vector.mul: invalid type of argument (number expected)";
}

