﻿Dimen = { };

Dimen._base = [
	{ symbol: 'L', unit: 'm' },
	{ symbol: 'T', unit: 's' },
	{ symbol: 'M', unit: 'g', prefix: 'k' },
	{ symbol: 'Θ', unit: 'K' },
	{ symbol: 'I', unit: 'A' },
	{ symbol: 'N', unit: 'mol' }
];

Dimen.formatUnit = function(dim) {
	var res = [];
	dim.forEach(function(p, index){
		if (p !== 0) {
			var u = Dimen._base[index].unit;
			if (Dimen._base[index].prefix)
				u = Dimen._base[index].prefix + u;
			if (p !== 1)
				u += '^' + p;
			res.push(u);		
		}
	});
	
	return res.join(' ');
}

Dimen._length    = [1, 0, 0, 0, 0, 0];
Dimen._time      = [0, 1, 0, 0, 0, 0];
Dimen._mass      = [0, 0, 1, 0, 0, 0];
Dimen._temp      = [0, 0, 0, 1, 0, 0];
Dimen._current   = [0, 0, 0, 0, 1, 0];
Dimen._substance = [0, 0, 0, 0, 0, 1];
Dimen._speed     = Dimen._length.sub(Dimen._time);
Dimen._acc       = Dimen._speed.sub(Dimen._time);
Dimen._momentum  = Dimen._mass.add(Dimen._speed);
Dimen._force     = Dimen._mass.add(Dimen._acc);
Dimen._energy    = Dimen._force.add(Dimen._length);
Dimen._power     = Dimen._energy.sub(Dimen._time);
Dimen._area      = Dimen._length.mul(2);
Dimen._volume    = Dimen._length.mul(3);
Dimen._density   = Dimen._mass.sub(Dimen._volume);
Dimen._pressure  = Dimen._force.sub(Dimen._area);
Dimen._freq      = Dimen._time.mul(-1);
Dimen._charge    = Dimen._current.add(Dimen._time);
Dimen._voltage   = Dimen._energy.sub(Dimen._charge);
Dimen._magflux   = Dimen._voltage.add(Dimen._time);
Dimen._magind    = Dimen._magflux.sub(Dimen._area);
Dimen._capacitance  = Dimen._charge.sub(Dimen._voltage);
Dimen._resistance   = Dimen._voltage.sub(Dimen._current);
Dimen._conductance  = Dimen._resistance.mul(-1);
Dimen._conductivity = Dimen._conductance.sub(Dimen._length);
Dimen._inductance   = Dimen._magflux.sub(Dimen._current);


Dimen.isDimen = function(x) {
	return Array.isArray(x) && x.length === Dimen._base.length && x.every(Number.isFinite);
}

Dimen.index = function(num) {
	if (!Number.isInteger(num) || num < 0 || num >= Dimen._base.length)
		throw "Dimen.index: invalid dimension index";
	
	var res = Dimen._base.map(Math.zero);
	res[num] = 1;
	return res;
};

function Unit(symbol, dim, value, canpre) {
	if (!(this instanceof Unit))
		return new Unit(symbol, dim, value);
	
	if (Number.isInteger(dim))
		dim = Dimen.index(dim);

	if (typeof value === 'undefined')
		value = 1;
	else if (typeof value === 'object')
		value = value.valueOf();

	if (typeof canpre === 'undefined')
		canpre = true;

	if (!String.isString(symbol) || !Array.isArray(dim) || !Number.isFinite(value) || typeof canpre !== "boolean")
		throw "Unit: invalid parameters";

	this._symbol = symbol;
	this._dim = dim;
	this._value = value;
	this._canpre = canpre; //can apply prefix
	return this;
}

//named SI units, can prefixed
Unit._SI = function(symbol, dim) {
	var u = Object.create(Unit.prototype);
	u._symbol = symbol;
	u._dim = dim;
	u._value = 1;
	return u;
}

//non-SI units, can prefixed
Unit._pre = function(symbol, dim, value) {
	var u = Object.create(Unit.prototype);
	u._symbol = symbol;
	u._dim = dim;
	u._value = value;
	u._canpre = true;
	return u;
}

//non-SI units, cannot prefixed
Unit._nopr = function(symbol, dim, value) {
	var u = Object.create(Unit.prototype);
	u._symbol = symbol;
	u._dim = dim;
	u._value = value;
	u._canpre = false;
	return u;
}

Unit.prototype.toString = function() {
	return this._symbol;
}

Unit._om = {
	T: 12,
	G: 9,
	M: 6,
	k: 3,
	d: -1,
	c: -2,
	m: -3,
	µ: -6,
	n: -9,
	p: -12
};

Unit._all = (function(){
	var list = Dimen._base.map(function(dim, index){
		return new Unit(dim.unit, index, dim.prefix ? Math.pow(10, -Unit._om[dim.prefix]) : 1);
	});
	
	list.push(Unit._SI('N', Dimen._force));
	list.push(Unit._SI('J', Dimen._energy));
	list.push(Unit._SI('W', Dimen._power));
	list.push(Unit._SI('Pa', Dimen._pressure));
	list.push(Unit._SI('C', Dimen._charge));
	list.push(Unit._SI('V', Dimen._voltage));
	list.push(Unit._SI('F', Dimen._capacitance));
	list.push(Unit._SI('Ω', Dimen._resistance));
	list.push(Unit._SI('ohm', Dimen._resistance));
	list.push(Unit._SI('S', Dimen._conductance));
	list.push(Unit._SI('Wb', Dimen._magflux));
	list.push(Unit._SI('T', Dimen._magind));
	list.push(Unit._SI('H', Dimen._inductance));
	list.push(Unit._SI('Hz', Dimen._freq));
	list.push(Unit._pre('L', Dimen._volume, 0.001));
	list.push(Unit._pre('eV', Dimen._energy, 1.6021766208e-19));
	list.push(Unit._pre('bar', Dimen._pressure, 1e5));
	list.push(Unit._pre('cal', Dimen._energy, 4.184));
	list.push(Unit._pre('erg', Dimen._energy, 1e-7));
	list.push(Unit._pre('dyn', Dimen._force, 1e-5));
	list.push(Unit._pre('pc', Dimen._length, 3.08567758149137e16));
	list.push(Unit._nopr('ha', Dimen._area, 1e4));
	list.push(Unit._nopr('barn', Dimen._area, 1e-28));
	list.push(Unit._nopr('in', Dimen._length, 0.0254));
	list.push(Unit._nopr('ft', Dimen._length, 0.3048));
	list.push(Unit._nopr('mile', Dimen._length, 1609.344));
	list.push(Unit._nopr('yd', Dimen._length, 0.9144));
	list.push(Unit._nopr('AU', Dimen._length, 149597870700));
	list.push(Unit._nopr('ly', Dimen._length, 9.4605284e15));
	list.push(Unit._nopr('y', Dimen._time, 31557600));
	list.push(Unit._nopr('d', Dimen._time, 86400));
	list.push(Unit._nopr('h', Dimen._time, 3600));
	list.push(Unit._nopr('min', Dimen._time, 60));
	list.push(Unit._nopr('lb', Dimen._mass, 0.45359237));
	list.push(Unit._nopr('oz', Dimen._mass, 0.028349523125));
	list.push(Unit._nopr('amu', Dimen._mass, 1.660539040e-27));
	list.push(Unit._nopr('kn', Dimen._speed, 0.514444444));
	list.push(Unit._nopr('c', Dimen._speed, 299792458));
	list.push(Unit._nopr('mph', Dimen._speed, 0.44704));
	list.push(Unit._nopr('kph', Dimen._speed, 0.277777777777777778));
	list.push(Unit._nopr('kgf', Dimen._force, 9.80665));
	list.push(Unit._nopr('lbf', Dimen._force, 4.448222));
	list.push(Unit._nopr('at', Dimen._pressure, 101325));
	list.push(Unit._nopr('mmHg', Dimen._pressure, 133.322387415));
	list.push(Unit._nopr('torr', Dimen._pressure, 133.322387415));
	list.push(Unit._nopr('psi', Dimen._pressure, 6894.757));
	list.push(Unit._nopr('hp', Dimen._power, 745.699872));
	
	return list;
})();

Unit.parse = function(text) {
	var value = 1;
	var dim = Dimen._base.map(Math.zero);
	var parts = text.split(' ');
	parts.forEach(function(part){
		var p = 1;
		var name = part;
		var i = part.indexOf('^');
		if (i >= 0) {
			p = Number(part.slice(i+1));
			name = part.slice(0, i);
			if (!Number.isInteger(p))
				throw "Unit.parse: unit '" + name + "' raised to fractional power";
		}
		
		var unit = Unit._all.find(function(u){
			return u._symbol === name;
		});
		
		if (typeof unit === 'undefined') {
			var prefix = name.charAt(0);
			var base = name.slice(1);
			if (prefix in Unit._om) {
				unit = Unit._all.find(function(u){
					return u._symbol === base;
				});
				
				if (typeof unit === 'undefined')
					throw "Unit.parse: unknown unit '" + name + "'";
			
				if (!unit._canpre)
					throw "Unit.parse: cannot apply prefix to the unit '" + base + "'";
			
				value *= Math.pow(10, Unit._om[prefix] * p);
			}
			else
				throw "Unit.parse: unknown unit '" + name + "'";
		}
		
		value *= Math.pow(unit._value, p);
		dim = dim.add(unit._dim.mul(p));
	});

	return {
		value: value,
		dim: dim
	};
}

function Qty(value, unit) {
	if (!(this instanceof Qty))
		return new Qty(value, unit);

	if (!Number.isNumber(value))
		throw "Qty: value must be a number";
	
	if (String.isString(unit)) {
		var obj = Unit.parse(unit);
		this._value = value * obj.value;
		this._dim = obj.dim;	
		this._orig = {value: value, unit: unit};
		return this;
	}
	
	if (Dimen.isDimen(unit)) {
		this._value = value;
		this._dim = unit;
		return this;
	}
	
	throw "Qty: invalid unit or dimension";
}

Qty._create = function(value, dim) {
	var q = Object.create(Qty.prototype);
	q._value = value;
	q._dim = dim;
	return q;
}

Qty.prototype.toStr = function(fn) {
	if (this._orig)
		return fn(this._orig.value) + ' ' + this._orig.unit;
		
	var res = fn(this._value);
	if (this.isNumber())
		return res;
	
	return res + ' ' + Dimen.formatUnit(this._dim);
}

Qty.prototype.isNumber = function() {
	return this._dim.every( function(n){ return n === 0; } );
}

Qty.prototype.add = function() {
	if (arguments.length === 2 && Number.isNumber(arguments[0]) && String.isString(arguments[1]))
		return this.add(new Qty(arguments[0], arguments[1]));

	if (arguments.length === 1 && arguments[0] instanceof Qty) {
		var x = arguments[0];
		if (!x._dim.equals(this._dim))
			throw "Qty.add: dimensions do not match";
			
		return Qty._create(this._value + x._value, this._dim);
	}
	
	if (arguments.length === 1 && Number.isNumber(arguments[0]))
		if (this.isNumber())
			return Qty._create(this._value + arguments[0], this._dim);
		else
			throw "Qty.add: dimensions do not match";

	throw "Qty.add: invalid arguments";
}

Qty.prototype.sub = function() {
	if (arguments.length === 2 && Number.isNumber(arguments[0]) && String.isString(arguments[1]))
		return this.sub(new Qty(arguments[0], arguments[1]));

	if (arguments.length === 1 && arguments[0] instanceof Qty) {
		var x = arguments[0];
		if (!x._dim.equals(this._dim))
			throw "Qty.sub: dimensions do not match";
			
		return Qty._create(this._value - x._value, this._dim);
	}
	
	if (arguments.length === 1 && Number.isNumber(arguments[0]) && this._dim.every(function(d){ return d === 0; }))
		return Qty._create(this._value - arguments[0], this._dim);

	throw "Qty.sub: invalid arguments";
}

Qty.prototype.mul = function() {
	if (arguments.length === 2 && Number.isNumber(arguments[0]) && String.isString(arguments[1]))
		return this.mul(new Qty(arguments[0], arguments[1]));

	if (arguments.length === 1 && arguments[0] instanceof Qty) {
		var x = arguments[0];
		return Qty._create(this._value * x._value, this._dim.add(x._dim));
	}
	
	if (arguments.length === 1 && Number.isNumber(arguments[0]))
		return Qty._create(this._value * arguments[0], this._dim);

	throw "Qty.mul: invalid arguments";
}

Qty.prototype.div = function() {
	if (arguments.length === 2 && Number.isNumber(arguments[0]) && String.isString(arguments[1]))
		return this.div(new Qty(arguments[0], arguments[1]));

	if (arguments.length === 1 && arguments[0] instanceof Qty) {
		var x = arguments[0];
		return Qty._create(this._value / x._value, this._dim.sub(x._dim));
	}
	
	if (arguments.length === 1 && Number.isNumber(arguments[0]))
		return Qty._create(this._value / arguments[0], this._dim);

	throw "Qty.div: invalid arguments";
}

Qty.prototype.pow = function(n) {
	if (!Number.isInteger(n)) 
		throw "Qty.pow: can only raise to an integral power";
	
	return Qty._create(Math.pow(this._value, n), this._dim.mul(n));
}

Qty.prototype.in = function(unit) {
	if (!String.isString(unit)) 
		throw "Qty.in: invalid argument (string expected)";
	
	var obj = Unit.parse(unit);
	if (!this._dim.equals(obj.dim))
		throw "Qty.in: dimensions do not match";
		
	return this._value / obj.value;
}

Qty.prototype.to = function(unit) {
	return new Qty(this.in(unit), unit);
}

Qty.prototype.exp = function() {
	if (!this.isNumber())
		throw "Qty.exp: can only apply to a dimensionless quantity";

	return Qty._create(Math.exp(this._value), this._dim);
}

Qty.prototype.dimen = function() {
	var res = [];
	
	this._dim.forEach(function(p, index){
		if (p !== 0)
			res.push(Dimen._base[index].symbol + (p !== 1 ? '^' + p : ''));
	});
	
	return res.join(' ');
}



